import time

import sourmash
from sourmash_plugin_branchwater import sourmash_plugin_branchwater as branch

from .database_info import MOLTYPE, KSIZE, SCALED


def load_sig(fullpath):
    fullpath = str(fullpath)
    try:
        ss = sourmash.load_file_as_index(fullpath)
        ss = ss.select(moltype=MOLTYPE, ksize=KSIZE, scaled=SCALED)
        if len(ss) == 1:
            ss = list(ss.signatures())[0]
            if len(ss.minhash):
                return ss
    except:
        pass

    return None


def run_gather(outpath, csv_filename, db_info):
    start = time.time()
    status = branch.do_fastmultigather(
        outpath,
        db_info.filename,
        0,
        KSIZE,
        SCALED,
        MOLTYPE,
        csv_filename,
        False,
        False,
    )
    end = time.time()

    print(f"branchwater gather status: {status}; time: {end - start:.2f}s")
    return status


def sig_is_assembly(ss):
    mh = ss.minhash
    # track abundance set? => assembly
    if not mh.track_abundance:
        print('ZZZ1 - is assembly')
        return True

    # count the number > 1 in abundance
    n_above_1 = sum(1 for (hv, ha) in mh.hashes.items() if ha > 1)
    f_above_1 = n_above_1/len(mh)
    print(f'n_above_1: {n_above_1}, of {len(mh)}, f={f_above_1:.3}')

    # more than 10% > 1? => probably not assembly
    if f_above_1 > 0.1:
        return False

    # nope! assembly!
    return True


def estimate_weight_of_unknown(ss, db, *, CUTOFF=5):
    mh = ss.minhash
    merged_hashes = db.merged_hashes

    unknown = [ (hv, ha) for (hv, ha) in mh.hashes.items() if ha not in merged_hashes ]
    sum_unknown = sum( ha for (hv, ha) in unknown )
    sum_high = sum( ha for (hv, ha) in unknown if ha >= CUTOFF )
    sum_low = sum( ha for (hv, ha) in unknown if ha < CUTOFF )

    return sum_high / sum_unknown, sum_low / sum_unknown




import os
import tempfile
import time
import gzip
import shutil
import glob

from flask import Flask, flash, request, redirect, url_for
from flask import render_template, send_from_directory
from werkzeug.utils import secure_filename
import markdown

import pandas as pd

from . import jinja2_filters
from .database_info import databases as sourmash_databases
from .database_info import MOLTYPE, KSIZE, SCALED
from .utils import *

default_settings = dict(
    UPLOAD_FOLDER = os.path.join(os.path.dirname(__file__), '../chill-data'),
    EXAMPLES_DIR = os.path.join(os.path.dirname(__file__), "examples/")
)


app = None
def init():
    global app

    app = Flask(__name__)
    app.config.update(default_settings)
    if 'CHILL_FILTER_SETTINGS' in os.environ:
        app.config.from_envvar('CHILL_FILTER_SETTINGS')

    jinja2_filters.add_filters(app.jinja_env.filters)

    try:
        os.mkdir(app.config['UPLOAD_FOLDER'])
    except FileExistsError:
        pass

    if 0:
        start = time.time()
        print(f'loading dbs:')
        for db in sourmash_databases:
            db.load()
            print(f'...done! {time.time() - start:.1f}s')


init()


def create_app():
    # Quick test configuration. Please use proper Flask configuration options
    # in production settings, and use a separate file or environment variables
    # to manage the secret key!
    app.secret_key = "super secret key"
    app.config["SESSION_TYPE"] = "filesystem"

    # sess.init_app(app)

    app.debug = True
    return app

###
### actual Web site stuff
###

@app.route("/", methods=["GET", "POST"])
def index():
    if request.method == "POST":
        # check if the post request has the file part
        if "sketch" not in request.files:
            flash("No file part")
            return redirect(request.url)
        file = request.files["sketch"]

        # If the user does not select a file, the browser submits an
        # empty file without a filename.
        if file.filename == "":
            flash("No selected file")
            return redirect(request.url)

        if file:
            filename = secure_filename(file.filename)
            outpath = os.path.join(app.config['UPLOAD_FOLDER'], filename)
            file.save(outpath)

            ss = load_sig(outpath)
            if ss:
                md5 = ss.md5sum()[:8]
                return redirect(f"/{md5}/{filename}/search")

    # default
    return render_template("index.html")


@app.route("/sketch", methods=["GET", "POST"])
def sketch():
    if request.method == "POST":
        # check if the post request has the file part
        if "signature" not in request.form:
            flash("No file part")
            return redirect(request.url)

        # take uploaded file and save
        sig_json = request.form["signature"]
        success = False
        filename = f"t{int(time.time())}.sig.gz"
        outpath = os.path.join(UPLOAD_FOLDER, filename)
        with gzip.open(outpath, "wt") as fp:
            fp.write(f"[{sig_json}]")

        ss = load_sig(outpath)
        if ss:
            # success? build URL & redirect
            md5 = ss.md5sum()[:8]
            return redirect(f"/{md5}/{filename}/search")

    # default: redirect to /
    return redirect(url_for("index"))


@app.route("/example", methods=["GET"])
def example():
    "Retrieve an example"
    filename = request.args["filename"]
    filename = secure_filename(filename)
    frompath = os.path.join(app.config['EXAMPLES_DIR'], filename)
    if not os.path.exists(frompath):
        return render_template("error.html",
                               error_message=f"example file <tt>{filename}</tt> not found in examples"), 404

    ss = load_sig(frompath)
    if ss is None:
        # doesn't match moltype etc, or other problems.
        return render_template("error.html",
                               error_message="Internal error: bad example file!?"), 404

    md5 = ss.md5sum()[:8]

    # now build the filename & make sure it's in the upload dir.
    topath = os.path.join(app.config['UPLOAD_FOLDER'], filename)
    if not os.path.exists(topath):
        print(f"copying: {frompath} {topath}")
        shutil.copy(frompath, topath)

    return redirect(f"/{md5}/{filename}/search")


@app.route("/")
@app.route("/<path:path>")
def get_by_md5(path):
    print("PATH IS:", path, os.path.split(path))
    path = path.split("/")
    if len(path) != 3:
        return redirect(url_for("index"))

    md5, filename, action = path

    # now try loading the sketch
    sigpath = os.path.join(app.config['UPLOAD_FOLDER'], filename)
    success = False
    ss = None
    if os.path.exists(sigpath):
        ss = load_sig(sigpath)
        if ss and ss.md5sum()[:8] == md5:
            success = True

    print('SUCCESS VALUE:', success)
    if not success:
        return redirect(url_for("index"))

    assert ss is not None
    sample_name = ss.name or "(unnamed sample)"

    # actions!
    if action == 'download_csv':
        csv_filename = filename + ".x.all.gather.csv" # @CTB
        return send_from_directory(app.config['UPLOAD_FOLDER'], csv_filename)
    elif action == "download":
        return send_from_directory(app.config['UPLOAD_FOLDER'], filename)
    elif action == "delete":
        file_list = glob.glob(f"{sigpath}.*.csv")
        for filename in file_list + [sigpath]:
            try:
                print('removing:', (filename,))
                os.unlink(filename)
            except:
                pass
        return redirect(url_for("index"))
    elif action == "search":
        search_db = None
        for db in sourmash_databases:
            if db.default:
                search_db = db
                break

        if search_db is None:
            raise Exception("no default search DB?!")

        csv_filename = f"{sigpath}.x.{search_db.shortname}.gather.csv"
        if not os.path.exists(csv_filename):
            status = run_gather(sigpath, csv_filename, search_db)
            if status != 0:
                return "search failed, for reasons that are probably not your fault"
            else:
                print(f'output is in: "{csv_filename}"')
        else:
            print(f"using cached output in: '{csv_filename}'")

        # read!
        try:
            gather_df = pd.read_csv(csv_filename)
            gather_df = gather_df[gather_df["f_unique_weighted"] >= 0.001]
        except:
            gather_df = []

        # no (significant) results?? exit.
        if not len(gather_df):
            return render_template(
                "sample_search_no_matches.html",
                sample_name=sample_name)

        # ok, now prep for display.

        gather_df['match_description'] = gather_df['match_name'].apply(search_db.get_display_name)

        # process abundance-weighted matches
        if not sig_is_assembly(ss):
            #f_unknown_high, f_unknown_low = estimate_weight_of_unknown(ss,
            #         search_db)
            f_unknown_high, f_unknown_low = 0, 0

            last_row = gather_df.tail(1).squeeze()
            sum_weighted_found = last_row["sum_weighted_found"]
            total_weighted_hashes = last_row["total_weighted_hashes"]

            f_found = sum_weighted_found / total_weighted_hashes

            return render_template(
                "sample_search_abund.html",
                sample_name=sample_name,
                sig=ss,
                gather_df=gather_df,
                f_found=f_found,
                f_unknown_high=f_unknown_high,
                f_unknown_low=f_unknown_low,
            )
        # process flat matching (assembly)
        else:
            print('running flat')
            last_row = gather_df.tail(1).squeeze()
            f_found = gather_df['f_unique_to_query'].sum()

            return render_template(
                "sample_search_flat.html",
                sample_name=sample_name,
                sig=ss,
                gather_df=gather_df,
                f_found=f_found,
            )

    # default: sample index
    sum_weighted_hashes = sum(ss.minhash.hashes.values())
    return render_template(
        "sample_index.html",
        sig=ss,
        sig_filename=filename,
        sample_name=sample_name,
        sum_weighted_hashes=sum_weighted_hashes,
    )

@app.route("/faq")
def faq():
    return render_template("faq.md")

@app.route("/guide")
def guide():
    return render_template("guide.md")

@app.route("/favicon.ico")
def favicon():
    return ""



from flask import Flask, session
import argparse

# from flask.ext.session import Session

SESSION_TYPE = "memcache"

if __name__ == "__main__":
    from . import create_app

    p = argparse.ArgumentParser()
    p.add_argument("-p", "--port", default=5000, type=int)
    p.add_argument("--host", default="127.0.0.1")
    args = p.parse_args()
    app = create_app()
    app.run(port=args.port, host=args.host)



# project-specific filters for jinja2 templating
import markdown

def markdownify(md):
    return markdown.markdown(md)

def percent(value):
    return f"{value*100:.1f}%"

def format_bp(bp):
    "Pretty-print bp information."
    bp = float(bp)
    if bp < 500:
        return f"{bp:.0f} bp"
    elif bp <= 500e3:
        return f"{round(bp / 1e3, 1):.1f} kbp"
    elif bp < 500e6:
        return f"{round(bp / 1e6, 1):.1f} Mbp"
    elif bp < 500e9:
        return f"{round(bp / 1e9, 1):.1f} Gbp"
    return f"??? {bp}"

def unique_weighted_bp(item):
    return format_bp(item['n_unique_weighted_found'] * item['scaled'])

def unique_flat_bp(item):
    return format_bp(item['unique_intersect_bp'])

filters_dict = {}
filters_dict['markdownify'] = markdownify
filters_dict['percent'] = percent
filters_dict['format_bp'] = format_bp
filters_dict['unique_weighted_bp'] = unique_weighted_bp
filters_dict['unique_flat_bp'] = unique_flat_bp

def add_filters(env_d):
    for k, v in filters_dict.items():
        env_d[k] = v



import os
from chill_filter_web import utils


def test_front(client):
    response = client.get("/")
    print(response.data)
    assert b"chill-filter: What's in my sample?" in response.data


def test_guide(client):
    response = client.get("/guide")
    assert b"The chill-filter user guide" in response.data


def test_faq(client):
    response = client.get("/faq")
    assert b"Frequently Asked Questions - chill-filter" in response.data


def test_upload_d_exists(app):
    # this is really a test for the test fixtures ;)
    with app.app_context():
        dirpath = app.config['UPLOAD_FOLDER']
        print(dirpath)
        assert os.path.exists(os.path.join(dirpath, 'Bu5.abund.k51.s100_000.sig.zip.x.all.gather.csv'))


def test_fail_example(client):
    response = client.get('/example?filename=XXX_Bu5.abund.k51.s100_000.sig.zip')
    print(response.data)
    assert response.status_code == 404


def test_bad_example(client):
    response = client.get('/example?filename=bad-example.sig.zip')
    print(response.data)
    assert response.status_code == 404


def test_display_abund(client):
    response = client.get('/example?filename=Bu5.abund.k51.s100_000.sig.zip',
                          follow_redirects=True)
    print(response.data)
    assert b'this looks like a set of reads' in response.data
    assert b'at least <b>81.5%</b> of your sequencing data will map' in response.data
    assert b'76.3% (60.1 Gbp)' in response.data
    assert b'4.8% (3.8 Gbp)' in response.data
    assert b'0.3% (215.2 Mbp)' in response.data
    assert b'bacteria and archaea (GTDB rs220)' in response.data
    assert b'22x' in response.data
    assert b'1x' in response.data    


def test_display_flat(client):
    response = client.get('/example?filename=Bu5.flat.k51.s100_000.sig.zip',
                          follow_redirects=True)
    print(response.data)
    assert b'this looks like an assembly' in response.data
    assert b'<b>90.7%</b>\n  of your contigs' in response.data
    assert b'83.5% (2.1 Gbp)' in response.data
    assert b'bacteria and archaea (GTDB rs220) ' in response.data
    assert b'7.1% (176.1 Mbp)' in response.data


def test_display_no_matches(client):
    response = client.get('/ac5b62eb/no-matches.sig.zip/search',
                          follow_redirects=True)
    print(response.data)
    assert b'no matches to your sample' in response.data


def test_delete(app):
    # this is really a test for the test fixtures ;)
    with app.app_context():
        client = app.test_client()

        # copy the example over
        response = client.get('/example?filename=Bu5.abund.k51.s100_000.sig.zip',
                              follow_redirects=True)
        dirpath = app.config['UPLOAD_FOLDER']

        # the CSV should also exist
        assert os.path.exists(os.path.join(dirpath, 'Bu5.abund.k51.s100_000.sig.zip'))
        assert os.path.exists(os.path.join(dirpath, 'Bu5.abund.k51.s100_000.sig.zip.x.all.gather.csv'))
        response = client.get('/97681062/Bu5.abund.k51.s100_000.sig.zip/delete')
        assert not os.path.exists(os.path.join(dirpath, 'Bu5.abund.k51.s100_000.sig.zip'))
        assert not os.path.exists(os.path.join(dirpath, 'Bu5.abund.k51.s100_000.sig.zip.x.all.gather.csv'))


def test_sample_index(client):
    # copy it in...
    response = client.get('/example?filename=Bu5.abund.k51.s100_000.sig.zip',
                          follow_redirects=True)

    # ...check location.
    response = client.get('97681062/Bu5.abund.k51.s100_000.sig.zip/')
    assert b'Bu5.abund.k51.s100_000.sig.zip' in response.data


def test_sample_download_sketch(client, tmp_path):
    response = client.get('/example?filename=Bu5.abund.k51.s100_000.sig.zip',
                          follow_redirects=True)
    response = client.get('97681062/Bu5.abund.k51.s100_000.sig.zip/download')
    data = response.get_data()
    outfile = tmp_path / 'xyz.sig.zip'
    with open(outfile, 'wb') as fp:
        fp.write(data)

    assert utils.load_sig(outfile)


def test_sample_download_csv(client):
    response = client.get('/example?filename=Bu5.abund.k51.s100_000.sig.zip',
                          follow_redirects=True)
    response = client.get('97681062/Bu5.abund.k51.s100_000.sig.zip/download_csv')
    assert response.status_code == 200
    data = response.get_data()
    print(data)
    assert data.startswith(b'intersect_bp,f_orig_query,f_match,')





import sourmash

MOLTYPE = "DNA"
KSIZE = 51
SCALED = 100_000


class DatabaseDescription:
    def __init__(self, shortname, filename,
                 merged_sketch_filename,
                 sketch_to_display_names,
                 sketch_to_nextlevel,
                 *,
                 default=False):
        self.shortname = shortname
        self.filename = filename
        self.merged_sketch_filename = merged_sketch_filename
        self.sketch_to_display_names = sketch_to_display_names
        self.sketch_to_nextlevel = sketch_to_nextlevel
        self.default = default

    def load(self):
        ss = sourmash.load_file_as_index(self.merged_sketch_filename)
        ss = ss.select(moltype=MOLTYPE, ksize=KSIZE, scaled=SCALED)
        if len(ss) != 1:
            raise Exception("more than one sketch in merged sketches?!")
        
        ss = list(ss.signatures())[0]
        self.merged_hashes = ss.minhash.hashes

    def get_display_name(self, x):
        return self.sketch_to_display_names.get(x, x)

databases = [
    DatabaseDescription('all',
                        'prepare-db/plants+animals+gtdb.rocksdb',
                        'prepare-db/plants+animals+gtdb.merged.sig.zip',
                        {'bosTau9': 'cattle genome (bosTau9)',
                         'canFam6': 'dog genome (canFam6)',
                         'equCab3': 'horse genome (equCab3)',
                         'felCat9': 'cat genome (felCat9)',
                         'galGal6': 'chick genome (galGal6)',
                         'hg38+alt': 'human genome (hg38 + alt tracks)',
                         'mm39': 'mouse genome (mm39)',
                         'oviAri4': 'sheep genome (oviAri4)',
                         'susScr11': 'pig genome (suScr11)',
                         'genbank-plants': 'all plants (GenBank 12/2024)',
                         }, {}, default=True),
    ]
